<?php
/**
 * @file
 * Provide review functions.
 *
 * @author Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

/**
 * Run the review.
 */
function pifr_client_review_run() {
  if (PIFR_CLIENT_TEST_START) {
    watchdog('pifr_client', 'Attempt to start review while already running.', array(), WATCHDOG_ALERT);
    return;
  }
  if (!variable_get('pifr_client_test', FALSE)) {
    watchdog('pifr_client', 'Attempt to start review without a test.', array(), WATCHDOG_ERROR);
    return;
  }

  // Load test information.
  $test = variable_get('pifr_client_test', FALSE);

  // Keep track of how long review has been running.
  variable_set('pifr_client_test_start', time());

  $review = pifr_client_review_load($test);
  if ($review) {
    $review->run();
  }

  // Reset variables.
  module_load_include('cron.inc', 'pifr_client');
  pifr_client_cron_clear_variables();

  // Allow for reporting to be disabled by addon modules.
  if (!defined('PIFR_CLIENT_REPORT') || PIFR_CLIENT_REPORT) {
    // Report results.
    module_load_include('xmlrpc.inc', 'pifr_client');
    if ($review) {
      pifr_client_xmlrpc_send_result($review->get_result());
    }

    if (PIFR_ACTIVE) {
      // Request next test.
      pifr_client_cron_request_next();
    }
  }
}

/**
 * Load review base class and plugin specific class.
 *
 * @param array $test Test information.
 * @param $return_class Return the class name instead of instance.
 * @return object Instantiated review class, or FALSE if invalid review plugin.
 */
function pifr_client_review_load(array $test, $return_class = FALSE) {
  // Load base review class.
  require_once drupal_get_path('module', 'pifr') . '/review/client.inc';

  // Load review plugin and run review.
  if (module_exists($test['review']['plugin'])) {
    require_once drupal_get_path('module', $test['review']['plugin']) . '/' . $test['review']['plugin'] . '.client.inc';
    $class = 'pifr_client_review_' . $test['review']['plugin'];

    if ($return_class) {
      return $class;
    }

    // Initialize review.
    return new $class($test);
  }
  else {
    watchdog('pifr_client', 'Review plugin module not found [@plugin].', array('@plugin' => $test['review']['plugin']), WATCHDOG_ERROR);
  }
  return FALSE;
}

/**
 * Get the review database type.
 *
 * @return string Review database type, either: mysql, pgsql, or sqlite, or
 *   FALSE if it could not be determined.
 */
function pifr_client_review_database_type() {
  global $db_url;
  static $type = FALSE;

  if (!$type) {
    $type = parse_url($db_url['pifr_checkout'], PHP_URL_SCHEME);
    foreach (array('mysql', 'pgsql', 'sqlite') as $backend) {
      if (strpos($type, $backend) !== FALSE) {
        $type = $backend;
        break;
      }
    }
  }
  return $type;
}

/**
 * Client review form.
 */
function pifr_client_review_form(&$form_state) {
  $form = array();

  $form['run'] = array(
    '#type' => 'fieldset',
    '#title' => t('Run'),
  );
  $form['run']['type'] = array(
    '#type' => 'select',
    '#title' => t('Type'),
    '#description' => t('Full suite will be effected by the debug setting which can be set manually or auto ' .
      'triggered by the presence of SimpleTest.'),
    '#options' => array(
      'default' => t('Default - Drupal 7, full suite'),
      'quick' => t('Quick - Drupal 7, NonDefaultBlockAdmin test case'),
      'fail' => t('Fail - Drupal 7, full suite, PHP exceptions'),
      'd6' => t('Drupal 6 - Several dependencies, full suite'),
    ),
  );
  $form['run']['op'] = array(
    '#type' => 'submit',
    '#value' => t('Run'),
  );

  return $form;
}

/**
 * Run the pre-built tests.
 */
function pifr_client_review_form_submit($form, &$form_state) {
  // Set the test variable so that the admin interface reacts accordingly.
  $test = pifr_client_review_pre_built_test($form_state['values']['type']);
  variable_set('pifr_client_test', $test);

  if ($review = pifr_client_review_load($test)) {
    $review->run();

    // Display test information.
    drupal_set_message('Test: ' . pifr_client_review_export($test));
    drupal_set_message('Results: ' . pifr_client_review_export($review->get_result()));

    variable_del('pifr_client_test');
  }
}

/**
 * Get a pre-built test, given the $type.
 *
 * @param string $type Test type, 'default', 'quick', 'fail', 'd6'.
 * @return array Pre-built test information array.
 */
function pifr_client_review_pre_built_test($type) {
  // Get the default arguments to be passed to client.
  module_load_include('review.inc', 'pifr_server');
  $plugin = pifr_server_review_plugin_load('pifr_simpletest');
  $default = $plugin->argument_default_get();

  if ($type == 'default' || $type == 'quick' || $type == 'fail') {
    // Provide Drupal HEAD repository and blank patch.
    $test = array(
      'test_id' => 17,
      'review' => array(
        'plugin' => 'pifr_simpletest',
        'argument' => $default,
      ),
      'vcs' => array(
        'main' => array(
          'repository' => array(
            'type' => 'git',
            // GITMIGRATION Fragile
            'url' => variable_get('pifr_drupal_repository', 'git://git.drupal.org/project/drupal.git'),
          ),
          'vcs_identifier' => '7.x',
        ),
        'dependencies' => array(),
      ),
      'files' => array(
        url(drupal_get_path('module', 'pifr_simpletest') . '/confirmation/pass.patch', array('absolute' => TRUE)),
      ),
    );

    // Provide quickest test.
    if ($type == 'quick') {
      $test['review']['argument']['tests'] = array('NonDefaultBlockAdmin');
    }
    elseif ($type == 'fail') {
      $test['files'] = array(
        url(drupal_get_path('module', 'pifr_simpletest') . '/confirmation/fail.patch', array('absolute' => TRUE)),
      );
    }
  }
  elseif ($type == 'd6') {
    // Contrib test with dependencies.
    $test = array(
      'test_id' => 17,
      'review' => array(
        'plugin' => 'pifr_simpletest',
        'argument' => array(
          'drupal.core.version' => 6,
        ) + $default,
      ),
      'vcs' => array(
        'main' => array(
          'repository' => array(
            'type' => 'git',
            'url' => variable_get('pifr_drupal_repository', 'git://git.drupal.org/project/drupal.git'),
          ),
          'vcs_identifier' => '6.x',  // GITMIGRATION Fragile
        ),
        'dependencies' => array(
          array(
            'repository' => array(
              'type' => 'git',
              'url' => variable_get('pifr_multivariate_repository', 'git://git.drupal.org/project/multivariate.git'),
            ),
            'vcs_identifier' => 'master',  // GITMIGRATION Fragile
          ),
          array(
            'repository' => array(
              'type' => 'git',
              'url' => variable_get('pifr_chart_repository', 'git://git.drupal.org/project/chart.git'),
            ),
            'vcs_identifier' => 'master',  // GITMIGRATION Fragile
          ),
        ),
      ),
      'files' => array(
        url(drupal_get_path('module', 'pifr_simpletest') . '/confirmation/pass.patch', array('absolute' => TRUE)),
      ),
    );
  }
  return $test;
}

/**
 * Export a variable.
 *
 * @param mixed $var Variable to print.
 * @return string Exported variables.
 */
function pifr_client_review_export($var) {
  return '<div>' . highlight_string("<?php\n" . var_export($var, TRUE) . "\n?>", TRUE) . '</div>';
}
